import coincurve
from Crypto.Hash import keccak
import binascii

from src.pwrpy.pwrapisdk import PWRPY
from src.pwrpy.signer import Signature


class WalletResponse:
    def __init__(self, success, txnHash, error=None):
        self.success = success
        self.txnHash = txnHash
        self.error = error


class PWRWallet:
    def __init__(self, private_key_hex=None):
        if private_key_hex:
            self.private_key = coincurve.PrivateKey.from_hex(private_key_hex)
        else:
            self.private_key = coincurve.PrivateKey()
        self.public_key = self.private_key.public_key

    def get_private_key(self):
        return self.private_key.to_hex()

    def get_public_key(self):
        public_hex = self.public_key.format(compressed=False).hex()[2:]
        return public_hex

    def get_address(self):
        public_hex = self.get_public_key()
        keccak_hash = keccak.new(digest_bits=256)
        keccak_hash.update(bytes.fromhex(public_hex))
        address_bytes = keccak_hash.digest()[-20:]
        return "0x" + address_bytes.hex()

    def get_nonce(self):
        nonce = PWRPY.get_nonce_of_address(self.get_address())
        if not nonce.success:
            raise RuntimeError(nonce.message)
        return nonce.data

    def get_balance(self):
        balance_result = PWRPY.get_balance_of_address(self.get_address())
        if not balance_result.success:
            raise RuntimeError(balance_result.message)
        return balance_result.data

    def __create_wallet_response(self, response, final_txn):
        if response.success:

            txn_hash = Signature.create_tx_hash_hex(final_txn).hex()
            return WalletResponse(True, "0x" + txn_hash)
        else:
            return WalletResponse(False, None, response.message)

    def transfer_pwr(self, to, amount, nonce=None):
        if nonce is None:
            nonce_response = PWRPY.get_nonce_of_address(self.get_address())
            if not nonce_response.success:
                return WalletResponse(False, None, nonce_response.message)
            nonce = nonce_response.data

        if len(to.strip()) != 42:
            raise RuntimeError("Invalid address")
        if amount < 0:
            raise RuntimeError("Amount cannot be negative")
        if nonce < 0:
            raise RuntimeError("Nonce cannot be negative")
        # if amount + (98 * self.PWRJ.getFeePerByte()) > self.getBalance():
        #     raise RuntimeError("Insufficient balance")
        # if nonce < self.getNonce():
        #     raise RuntimeError("Nonce is too low")

        buffer = bytearray(33)
        buffer[0] = 0
        buffer[1:5] = nonce.to_bytes(4, byteorder='big')
        buffer[5:13] = amount.to_bytes(8, byteorder='big')
        buffer[13:] = binascii.unhexlify(to[2:])
        txn = bytes(buffer)
        signature = Signature.sign_message(self.private_key, txn)

        final_txn = bytearray(98)
        final_txn[:33] = txn
        final_txn[33:] = signature

        response = PWRPY.broadcast_txn(final_txn)
        return self.__create_wallet_response(response, final_txn)

    def send_vm_data_txn(self, vmId, data, nonce=None):
        if nonce is None:
            nonce_response = PWRPY.get_nonce_of_address(self.get_address())
            if not nonce_response.success:
                return WalletResponse(False, None, nonce_response.message)
            nonce = nonce_response.data

        if nonce < 0:
            raise RuntimeError("Nonce cannot be negative")
        if nonce < self.get_nonce():
            raise RuntimeError("Nonce is too low")

        data_len = len(data)

        buffer = bytearray(13 + data_len)
        buffer[0] = 5
        buffer[1:5] = nonce.to_bytes(4, byteorder='big')
        buffer[5:13] = vmId.to_bytes(8, byteorder='big')
        buffer[13:] = data
        txn = bytes(buffer)
        signature = Signature.sign_message(self.private_key, txn)

        txn_len = len(txn)

        final_txn = bytearray(13 + 65 + data_len)
        final_txn[:txn_len] = txn
        final_txn[txn_len:] = signature

        response = PWRPY.broadcast_txn(final_txn)
        return self.__create_wallet_response(response, final_txn)

    def delegate(self, to, amount, nonce=None):
        if nonce is None:
            nonce_response = PWRPY.get_nonce_of_address(self.get_address())
            if not nonce_response.success:
                return WalletResponse(False, None, nonce_response.message)
            nonce = nonce_response.data

        buffer = bytearray(33)
        buffer[0] = 3
        buffer[1:5] = nonce.to_bytes(4, byteorder='big')
        buffer[5:13] = amount.to_bytes(8, byteorder='big')
        buffer[13:] = binascii.unhexlify(to[2:])
        txn = bytes(buffer)
        signature = Signature.sign_message(self.private_key, txn)

        txn_len = len(txn)

        final_txn = bytearray(txn_len + 65)
        final_txn[:txn_len] = txn
        final_txn[txn_len:] = signature

        response = PWRPY.broadcast_txn(final_txn)
        return self.__create_wallet_response(response, final_txn)

    def withdraw(self, from_wallet, shares_amount, nonce=None):
        if nonce is None:
            nonce_response = PWRPY.get_nonce_of_address(self.get_address())
            if not nonce_response.success:
                return WalletResponse(False, None, nonce_response.message)
            nonce = nonce_response.data

        buffer = bytearray(33)
        buffer[0] = 4
        buffer[1:5] = nonce.to_bytes(4, byteorder='big')
        buffer[5:13] = shares_amount.to_bytes(8, byteorder='big')
        buffer[13:] = binascii.unhexlify(from_wallet[2:])
        txn = bytes(buffer)
        signature = Signature.sign_message(self.private_key, txn)

        txn_len = len(txn)

        final_txn = bytearray(txn_len + 65)
        final_txn[:txn_len] = txn
        final_txn[txn_len:] = signature

        response = PWRPY.broadcast_txn(final_txn)
        return self.__create_wallet_response(response, final_txn)

    def claim_vm_id(self, vm_id, nonce=None):
        if nonce is None:
            nonce_response = PWRPY.get_nonce_of_address(self.get_address())
            if not nonce_response.success:
                return WalletResponse(False, None, nonce_response.message)
            nonce = nonce_response.data

        buffer = bytearray(13)
        buffer[0] = 6
        buffer[1:5] = nonce.to_bytes(4, byteorder='big')
        buffer[5:13] = vm_id.to_bytes(8, byteorder='big')
        txn = bytes(buffer)
        signature = Signature.sign_message(self.private_key, txn)

        txn_len = len(txn)

        final_txn = bytearray(txn_len + 65)
        final_txn[:txn_len] = txn
        final_txn[txn_len:] = signature

        response = PWRPY.broadcast_txn(final_txn)
        return self.__create_wallet_response(response, final_txn)



class Delegator:
    def __init__(self, address, validator_address, shares, delegated_pwr):
        self._address = address
        self._validator_address = validator_address
        self._shares = shares
        self._delegated_pwr = delegated_pwr

    @property
    def address(self):
        return self._address

    @property
    def validator_address(self):
        return self._validator_address

    @property
    def shares(self):
        return self._shares

    @property
    def delegated_pwr(self):
        return self._delegated_pwr



import hashlib
import os
import coincurve
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import utils
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives import hashes
from Crypto.Hash import keccak

from dotenv import load_dotenv

from pwrpy.pwrwallet import PWRWallet


load_dotenv(".env")


PRIVATE_KEY_HEX = os.environ.get("PRIVATE_KEY_HEX")


wallet = PWRWallet(PRIVATE_KEY_HEX)
target_address = "0xe744afead6a2115a3515506868840681c29f67e0"

print("Private Key:", wallet.get_private_key())
print("Public Key:", wallet.get_public_key())
print("Ethereum Address:", wallet.get_address())

nonce = wallet.get_nonce()

r = wallet.get_balance()

print(r)



import os
import requests
import json
from binascii import hexlify

from pwrpy.models.Block import Block
from pwrpy.models.Transactions import Transaction

if os.environ.get("PRC_NODE_URL") is None:
    raise RuntimeError("Please set the PRC_NODE_URL environment variable")


class ApiResponse:
    def __init__(self, success, message, data=None):
        self.success = success
        self.message = message
        self.data = data


class PWRPY:
    # Replace with the actual RPC node URL
    __PRC_NODE_URL = os.environ.get("PRC_NODE_URL")

    @staticmethod
    def get_rpc_node_url():
        return PWRPY.__PRC_NODE_URL

    @staticmethod
    def get_fee_per_byte():
        return 100

    @staticmethod
    def broadcast_txn(txn):
        try:
            url = PWRPY.__PRC_NODE_URL + "/broadcast/"
            headers = {
                "Accept": "application/json",
                "Content-type": "application/json"
            }
            payload = json.dumps({"txn": hexlify(txn).decode()})

            responseRaw = requests.post(url, data=payload, headers=headers)

            response = responseRaw.json()
            is_ok = responseRaw.status_code == 200

            return ApiResponse(is_ok, response.get("message"))

        except Exception as e:
            return ApiResponse(False, str(e))

    @staticmethod
    def get_nonce_of_address(address):
        try:
            url = PWRPY.__PRC_NODE_URL + "/nonceOfUser/?userAddress=" + address
            headers = {
                "Accept": "application/json",
                "Content-type": "application/json"
            }

            responseRaw = requests.get(url, headers=headers)

            response = responseRaw.json()

            if responseRaw.status_code != 200:
                return ApiResponse(False, response.get("message"))
            else:
                return ApiResponse(True, response.get("message"), response.get("nonce"))

        except Exception as e:
            return ApiResponse(False, str(e))

    @staticmethod
    def get_balance_of_address(address):
        try:
            url = PWRPY.__PRC_NODE_URL + "/balanceOf/?userAddress=" + address
            headers = {
                "Accept": "application/json",
                "Content-type": "application/json"
            }

            responseRaw = requests.get(url, headers=headers)

            response = responseRaw.json()

            if responseRaw.status_code != 200:
                return ApiResponse(False, response.get("message"))
            else:
                return ApiResponse(True, response.get("message"), response.get("balance"))

        except Exception as e:
            return ApiResponse(False, str(e))

    @staticmethod
    def get_blocks_count():
        try:
            url = PWRPY.__PRC_NODE_URL + "/blocksCount/"
            headers = {
                "Accept": "application/json",
                "Content-type": "application/json"
            }

            responseRaw = requests.get(url, headers=headers)

            response = responseRaw.json()

            if responseRaw.status_code != 200:
                return ApiResponse(False, response.get("message"))
            else:
                return ApiResponse(True, response.get("message"), response.get("blocksCount"))

        except Exception as e:
            return ApiResponse(False, str(e))

    @staticmethod
    def get_block_by_number(block_number):
        try:
            response = requests.get(
                f"{PWRPY.__PRC_NODE_URL}/block/?blockNumber={block_number}")

            if response.status_code == 200:
                json_data = response.json().get('block')
                # Assuming the Block constructor takes the JSON data directly

                block_instance = Block(
                    transaction_count=json_data['transactionCount'],
                    size=json_data['blockSize'],
                    number=json_data['blockNumber'],
                    reward=json_data['blockReward'],
                    timestamp=json_data['timestamp'],
                    hash=json_data['blockHash'],
                    submitter=json_data['blockSubmitter'],
                    success=json_data['success'],
                    transactions=[Transaction(size=txn['size'], hash=txn['hash'], fee=txn['fee'], from_address=txn['from'], to=txn['to'],
                                              nonce_or_validation_hash=txn['nonceOrValidationHash'], position_in_the_block=txn['positionInTheBlock'], type=txn['type']) for txn in json_data['transactions']]
                )

                return block_instance
            elif response.status_code == 400:
                error_data = response.json()
                raise RuntimeError(f"Failed with HTTP error 400 and message: {
                                   error_data.get('message')}")
            else:
                raise RuntimeError(f"Failed with HTTP error code : {
                                   response.status_code}")

        except requests.HTTPError as http_err:
            raise RuntimeError(f"HTTP error occurred: {http_err}")
        except Exception as err:
            raise RuntimeError(f"An error occurred: {err}")

    @staticmethod
    def get_latest_block_number():
        response = PWRPY.get_blocks_count()
        if not response.success:
            return ApiResponse(False, response.get("message"))

        return ApiResponse(True, None, response.data - 1)
